from pathlib import Path
import json

from k2s.conda import CondaManager


def _make_spec(path_to_python, display_name):
    """Create Jupyter kernel spec

    Parameters
    ----------
    path_to_python : str
        Path to bin/python in the target environment (the env must have
        ipykernel installed)

    display_name : str
        Name displayed to the user
    """
    return {
        "argv": [
            str(path_to_python), "-m", "ipykernel_launcher", "-f",
            "{connection_file}"
        ],
        "display_name":
        f"Python 3 ({display_name})",
        "language":
        "python",
        "metadata": {
            "debugger": True
        }
    }


def install(name=None,
            requirements=None,
            requirements_pip=None,
            verbose=False):
    """Create a new kernel
    """
    cm = CondaManager(install_conda=True)

    prefix_new_env = cm.create_env(
        name=name,
        requirements=requirements,
        requirements_pip=requirements_pip,
    )

    # register kernel if we created a new environment
    if name is not None:
        # newly created environment
        path_to_python = Path(prefix_new_env, 'bin', 'python')

        # folder where we'll register the kernel (in the prefix)
        # NOTE: I don't think this will work every time, we need to test with
        # sagemaker, google cloud notebooks, etc. - probably better to
        # install in the local kernels directory
        prefix = cm.get_active_prefix()
        path_to_kernels = Path(prefix, 'share', 'jupyter', 'kernels')

        spec = _make_spec(path_to_python, name)

        path_to_kernel = Path(path_to_kernels, name)
        path_to_kernel.mkdir(exist_ok=True)

        (path_to_kernel / 'kernel.json').write_text(json.dumps(spec))



import sys
import argparse

from k2s.bootstrap import from_url
from k2s.exceptions import CLIError


class CLI:

    def __init__(self) -> None:
        parser = argparse.ArgumentParser()
        parser.add_argument('command')
        args = parser.parse_args(sys.argv[1:2])

        if not hasattr(self, args.command):
            sys.exit(f'Unrecognized command {args.command!r}')
        else:
            cmd = getattr(self, args.command)

            try:
                cmd()
            except CLIError as e:
                sys.exit(e)

            sys.exit(0)

    def get(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('url')
        args = parser.parse_args(sys.argv[2:])
        from_url(args.url)



import urllib.request
from pathlib import Path
import json


class ChannelData:
    """Get conda-forge channel data
    """

    def __init__(self):
        self._home = Path('~', '.k2s').expanduser()
        self._data = self.load_channel_data()

    def load_channel_data(self):
        self._home.mkdir(exist_ok=True, parents=True)

        target = self._home / "conda-forge.json"

        if not target.exists():
            urllib.request.urlretrieve(
                "https://conda.anaconda.org/conda-forge/channeldata.json",
                target)

        data = json.loads(target.read_text())

        return data

    def pkg_exists(self, names):
        """Check if the packages exist in conda-forge
        """
        exist, not_exist = [], []

        for name in names:
            name_sanitized = name if '=' not in name else name.split('=')[0]

            if self._data['packages'].get(name_sanitized):
                # conda uses = insted of pip's ==
                if '==' in name:
                    name = name.replace('==', '=')

                exist.append(name)
            else:
                not_exist.append(name)

        return exist, not_exist



import os
import json
import sys
import shutil
from pathlib import Path
import urllib.request
import subprocess
from os import environ

from ploomber_core.telemetry.telemetry import Telemetry
from ploomber_core.config import Config

from k2s.subprocess import _run_command
from k2s.index import ChannelData
from k2s.exceptions import KernelRestartRequired

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from importlib_metadata import version

BASE_MINI = "https://repo.anaconda.com/miniconda"

# https://docs.python.org/3/library/sys.html#sys.platform
URLS = {
    'linux': f"{BASE_MINI}/Miniconda3-latest-Linux-x86_64.sh",
    'darwin': f"{BASE_MINI}/Miniconda3-latest-MacOSX-x86_64.sh",
}

MINI_3_7 = (f"{BASE_MINI}/Miniconda3-py37_4.12.0-Linux-x86_64.sh")

IS_KAGGLE = "KAGGLE_DOCKER_IMAGE" in environ
IS_COLAB = "COLAB_GPU" in environ

telemetry = Telemetry(
    api_key="phc_P9SpSeypyPwxrMdFn2edOOEooQioF2axppyEeDwtMSP",
    package_name="k2s",
    version=version("k2s"),
)


class CondaManager:
    """Manage conda installations

    Parameters
    ----------
    install_conda : bool, default=False
        Force installing at ~/.k2s/conda, even if there's an existing
        installation
    """

    def __init__(self, install_conda=True) -> None:
        self.pre_setup()

        self.install_conda = install_conda

        if not shutil.which("conda") and not install_conda:
            raise RuntimeError("conda is not installed, to install it "
                               "pass install_conda=True")

        has_local_conda = (self.get_config_directory() / 'conda').is_dir()

        if (not shutil.which('conda')
                or install_conda) and not has_local_conda:
            home = self.get_config_directory()
            self.install_conda_in_prefix(str(home / 'conda'))

        # ensure mamba is installed in the base prefix
        prefix = self.get_base_prefix()
        self.install_mamba_in_prefix(prefix)
        self.post_conda_install()

    def pre_setup(self):
        pass

    def post_conda_install(self):
        pass

    def _get_conda_url(self):
        url = URLS.get(sys.platform)

        if url is None:
            raise RuntimeError("Only Linux and Mac are supported")

        return url

    def install_conda_in_prefix(self, prefix):
        """Installs miniconda in the given prefix
        """
        print(f"Installing conda ({prefix!r})...")
        urllib.request.urlretrieve(self._get_conda_url(), "miniconda.sh")
        _run_command(["bash", "miniconda.sh", "-b", "-f", "-p", prefix])
        print("Finished installing conda.")

    def install_mamba_in_prefix(self, prefix):
        if not Path(prefix, "bin", "mamba").exists():
            print("Installing mamba...")
            _run_command([
                self.get_base_conda_bin(),
                "install",
                "mamba",
                "-c",
                "conda-forge",
                "-y",
                "--prefix",
                prefix,
            ])
            print("Done installing mamba.")
        else:
            print("mamba already installed...")

    def create_env(self, name, requirements, requirements_pip):
        """
        Notes
        -----
        Theres a chance conda/mamba determine to re-install Python, example:

        mamba create --name test python=3.8.13 -c defaults --yes
        conda activate test
        mamba install ploomber -c conda-forge --dry-run

        This will crash the environment, so we must pin Python


        Examples
        --------
        >>> from k2s.conda import CondaManager
        >>> cm = CondaManager()
        >>> cm.create_env("some-env", requirements=["python=3.8.13"],
        ... requirements_pip=None)
        >>> cm.create_env("some-env", requirements=["ploomber"],
        ... requirements_pip=None)
        """
        # TODO: if the env already exists, say "updating" instead of
        # "installing"
        print(
            "Installing dependencies, this might take a few minutes. "
            "Join our community while you wait: https://ploomber.io/community")

        spec = {
            "name":
            name,
            "channels": ["conda-forge", "default"],
            "dependencies":
            list(requirements) +
            ["pip", "ipykernel", "python",
             dict(pip=requirements_pip)],
        }

        Path("env.yml").write_text(json.dumps(spec))

        if name is None:
            prefix = str(Path(self.get_active_prefix()))
        else:
            prefix = str(Path(self.get_base_prefix(), "envs", name))

        # TODO: revert this change after executing the update?
        # if the environment exists, we must pin Python
        # should I pin ipykernel as well?
        if Path(prefix).exists():
            bin = str(Path(prefix, 'bin', 'python'))
            cmd = ("from sys import version_info as v"
                   "; print(f'{v.major}.{v.minor}.{v.micro}')")
            out = subprocess.run([bin, "-c", cmd],
                                 check=True,
                                 capture_output=True)
            version = out.stdout.decode().strip()
            Path(prefix, "conda-meta").mkdir(exist_ok=True)
            pinned_path = Path(prefix, "conda-meta", "pinned")
            if pinned_path.exists():
                pinned = pinned_path.read_text()
            else:
                pinned = ''

            pinned_path.write_text(f'{pinned}\npython=={version}')

        cmd = [
            self.get_base_mamba_bin(),
            "env",
            "update",
            "--file",
            "env.yml",
            "--prefix",
            prefix,
            # NOTE: --prune is broken since conda 4.4
            # https://github.com/conda/conda/issues/7279,
            # it doesn't do anything and we noticed it sometimes
            # breaks stuff
        ]

        _run_command(cmd)

        print("Done installing dependencies.")

        return prefix

    def get_config_directory(self):
        """Returns the directory to store k2s configuration
        """
        home = Path('~', '.k2s').expanduser()
        home.mkdir(exist_ok=True)
        return home

    def get_base_prefix(self):
        """Get base conda base prefix to use for all commands
        """
        if self.install_conda:
            return str(self.get_config_directory() / 'conda')
        else:
            # NOTE: is this the best way to retrieve this?
            # maybe sys.prefix?
            return str(Path(shutil.which("conda")).parent.parent)

    def get_base_conda_bin(self):
        return str(Path(self.get_base_prefix(), "bin", "conda"))

    def get_base_mamba_bin(self):
        return str(Path(self.get_base_prefix(), "bin", "mamba"))

    def get_active_prefix(self):

        out = subprocess.run([self.get_base_conda_bin(), 'info', '--json'],
                             check=True,
                             capture_output=True)

        active_prefix = json.loads(out.stdout.decode())['active_prefix']

        if not active_prefix:
            raise RuntimeError("Unable to determine current environment. "
                               "For help: https://ploomber.io/community")

        return active_prefix

    def prune():
        """Delete all environments and kernels
        """
        # also delete local conda installation
        pass


class ColabConfig(Config):
    install_requirements_argument: list = None

    def path(cls):
        # remove this once this is closed:
        # https://github.com/ploomber/core/issues/16
        path_to_cfg = Path('~/.k2s/colab/config.yaml').expanduser()
        path_to_cfg.parent.mkdir(exist_ok=True, parents=True)
        return path_to_cfg


class ColabCondaManager(CondaManager):

    def __init__(self, install_conda=True) -> None:
        self.pre_setup()

        if shutil.which("conda"):
            print("conda already installed...")
        else:
            self.install_conda_in_prefix(self.get_active_prefix())
            self.install_mamba_in_prefix(self.get_active_prefix())
            self.post_conda_install()

    def pre_setup(self):
        if not IS_COLAB:
            raise RuntimeError(
                f"{type(self).__name__} should only be used in Colab")

        # this is not added by default on Colab
        sys.path.insert(
            0, f"{self.get_active_prefix()}/lib/python3.7/site-packages")

    def post_conda_install(self):
        os.rename(sys.executable, f"{sys.executable}.colab")
        LD_LIBRARY_PATH = (f"{self.get_active_prefix()}/lib"
                           f":{os.environ.get('LD_LIBRARY_PATH', '')}")

        Path(sys.executable).write_text(f"""\
#!/bin/bash
export LD_LIBRARY_PATH={LD_LIBRARY_PATH}
{sys.executable}.colab -x $@
""")

        subprocess.run(["chmod", "+x", sys.executable], check=True)

        from IPython import get_ipython
        get_ipython().kernel.do_shutdown(restart=True)
        raise KernelRestartRequired

    def get_base_prefix(self):
        return "/usr/local"

    def get_active_prefix(self):
        return self.get_base_prefix()

    def _get_conda_url(self):
        return MINI_3_7


class KaggleCondaManager(CondaManager):

    def pre_setup(self):
        if not IS_KAGGLE:
            raise RuntimeError(
                f"{type(self).__name__} should only be used in Kaggle")

    def get_base_prefix(self):
        return "/opt/conda"

    def get_active_prefix(self):
        return self.get_base_prefix()


@telemetry.log_call('k2s-install')
def install(requirements):
    """Install packages
    """
    cfg = None

    if IS_COLAB:
        cfg = ColabConfig()

        if requirements == cfg.install_requirements_argument:
            print('Packages already installed...')
            # instantiate the manager since it'll configure the path for
            # Colab to work
            ColabCondaManager()
            return

        class_ = ColabCondaManager
    elif IS_KAGGLE:
        class_ = KaggleCondaManager
    else:
        class_ = CondaManager

    cm = class_()
    cd = ChannelData()
    conda, pip = cd.pkg_exists(requirements)
    cm.create_env(name=None, requirements=conda, requirements_pip=pip)

    if IS_COLAB:
        cfg.install_requirements_argument = list(requirements)



import sys
import threading
import itertools
import time


class Spinner:
    """
    https://stackoverflow.com/a/58174909/709975
    """

    def __init__(self, message="", delay=0.1):
        self.spinner = itertools.cycle(['-', '/', '|', '\\'])
        self.delay = delay
        self.busy = False
        self.spinner_visible = False
        sys.stdout.write(message)

    def write_next(self):
        with self._screen_lock:
            if not self.spinner_visible:
                sys.stdout.write(next(self.spinner))
                self.spinner_visible = True
                sys.stdout.flush()

    def remove_spinner(self, cleanup=False):
        with self._screen_lock:
            if self.spinner_visible:
                sys.stdout.write('\b')
                self.spinner_visible = False
                if cleanup:
                    sys.stdout.write(' ')
                    sys.stdout.write('\r')
                sys.stdout.flush()

    def spinner_task(self):
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def __enter__(self):
        self._screen_lock = threading.Lock()
        self.busy = True
        self.thread = threading.Thread(target=self.spinner_task)
        self.thread.start()

    def __exit__(self, exception, value, tb):
        self.busy = False
        self.remove_spinner(cleanup=True)



import json
import re
from pathlib import Path, PurePosixPath, PureWindowsPath
import urllib.request
from urllib.parse import urlsplit, urlunsplit
from urllib.error import HTTPError

import parso
from isort import place_module

# https://ipython.readthedocs.io/en/stable/config/extensions/index.html
from k2s.env import install
from k2s.index import ChannelData

_BUILT_IN_EXTENSIONS = {'autoreload', 'storemagic'}

_PACKAGE_MAPPING = {
    'sklearn': 'scikit-learn',
    'sql': 'jupysql',
}


# FIXME: downloaded files might depend on other files, so we have to download
# them as well
def download_files(source, url):
    files = local_files(source)

    split = urlsplit(url)
    split = split._replace(path=str(PurePosixPath(split.path).parent))
    base = urlunsplit(split)

    deps = set()

    # TODO: download in parallel
    for file in files:
        try:
            # TODO: if .py or .ipynb, there might be more packages to install
            urllib.request.urlretrieve(f'{base}/{file}', file)
            print(f'Downloaded filed used in notebook: {file}')

            if Path(file).suffix == '.py':
                deps_ = extract_imports_from_path_to_script(file)
                deps = deps | set(deps_)
            elif Path(file).suffix == '.ipynb':
                deps_ = extract_imports_from_path_to_notebook(file)
                deps = deps | set(deps_)

        except HTTPError as e:
            print("It appears the notebook is using "
                  f"a file named {file!r}, but downloading it failed: {e}")

    return deps


def _get_previous_leaf(leaf, *, levels):

    for _ in range(levels):
        leaf = leaf.get_previous_leaf()

        if leaf is None:
            return None

    return leaf


def local_files(source):
    """
    Given some source code, infer what files are used as inputs. For example,
    calls to .read, .read_* functions, or calls to pathlib.Path
    """
    # nbformat.read("notebook.ipynb") or jupytext.read("notebook.ipynb")
    # Path("file.txt")
    # pd.read_{fmt}("path.ext")
    # look for calls to these functions
    targets = {'read', 'Path'}

    # TODO: ploomber-specific: pipeline.yaml (DAGSpec.find())
    # or ploomber build

    # TODO: ignore things like s3://

    def evaluator(leaf):
        is_string = leaf.type == 'string'

        if not is_string:
            return False

        prev = leaf.get_previous_leaf()

        if prev.type == 'operator' and prev.value == '=':
            prev = _get_previous_leaf(leaf, levels=4)
        else:
            prev = _get_previous_leaf(leaf, levels=2)

        if prev is None:
            return is_string

        return prev.value in targets or (prev.value.startswith('read_')
                                         and prev.value != 'read_sql')

    return string_literals(source, evaluator=evaluator)


def _is_matplotlib_setting(leaf):
    """
    Determine if a string literal is a matplotlib's setting
    """

    def check(leaf):
        parent = leaf.parent
        grandparent = leaf.parent.parent

        left, right = parent.children[0].value, parent.children[-1].value
        # when doing:
        # import matplotlib as mpl
        # mpl.rcParams['figure.figsize'] = (12, 8)
        var_nested = grandparent.children[1].children[1].value
        # when doing:
        # from matplotlib import rcParams
        # rcParams['axes.xmargin'] = 0.1
        var = grandparent.children[0].value

        return left == '[' and right == ']' and (var_nested == 'rcParams'
                                                 or var == 'rcParams')

    try:
        return check(leaf)
    except Exception:
        return False


def paths(source, *, raw=False):
    """
    Given some source code, parse string literals that look like paths
    """

    def evaluator(leaf):
        is_string = leaf.type == 'string'

        if not is_string:
            return False

        value = leaf.value

        constructor = (PureWindowsPath
                       if ':' in value or '\\' in value else PurePosixPath)

        path = constructor(value[1:-1])

        return not _is_matplotlib_setting(leaf) and (len(path.parts) >= 2
                                                     or path.suffix)

    return string_literals(source, evaluator=evaluator, raw=raw)


def string_literals(source, *, evaluator=None, raw=False):
    """Find string literals in the source code

    Parameters
    ----------
    source : str
        Source code

    evaluator : callable
        A function that returns true if the leaf should be ocnsidered a
        candidate or not
    """
    if evaluator is None:

        def evaluator(leaf):
            return leaf.type == 'string'

    mod = parso.parse(source)

    leaf = mod.get_first_leaf()

    literals = []

    while leaf:
        if evaluator(leaf):
            if raw:
                value = leaf.value
            else:
                value = leaf.value[1:-1]

            literals.append(value)

        leaf = leaf.get_next_leaf()

    return set(literals)


def extract_from_plain_text(text):
    # NOTE: +, :, /, and . are also required since users may have something
    # like: pip install git+https://github.com/ploomber/ploomber
    # we ignore those requirements but we need to parse them
    matches = re.findall(r'pip install ([\w \-\+\:\/\.]+)', text)
    matches

    deps = set([
        item for sublist in [match.split() for match in matches]
        for item in sublist
        # ignore options passed to "pip install"
        if not item.startswith('-')
        # ignore git+URL and similar:
        # https://pip.pypa.io/en/stable/topics/vcs-support/
        if '+' not in item
    ])
    return list(deps - {'k2s'})


# NOTE: imports might also be from local files, so we should download them
# NOTE: we copied this from soorgeon
# https://github.com/ploomber/soorgeon/blob/cb16dc51f326cf21486bff86dfec03b4d1f5c2d7/src/soorgeon/definitions.py#L24
def packages_used(tree):
    """
    Return a list of the packages used, correcting for some packages whose
    module name does not match the PyPI package (e.g., sklearn -> scikit-learn)
    Returns None if fails to parse them
    """

    def flatten(elements):
        return [i for sub in elements for i in sub]

    def _extract_names(node):
        if hasattr(node, 'children'):
            return extract_names(node.children[0])
        else:
            return [node.value]

    def extract_names(import_):
        if import_.type == 'name':
            return [import_.value]
        elif import_.type in {'dotted_name', 'dotted_as_name'}:
            return [import_.children[0].value]

        second = import_.children[1]

        if second.type in {'dotted_name', 'dotted_as_name'}:
            return extract_names(second.children[0])
        elif second.type == 'dotted_as_names':
            # import a as something, b as another

            return flatten([
                _extract_names(node) for i, node in enumerate(second.children)
                if i % 2 == 0
            ])
        else:
            return [second.value]

    pkgs = flatten([extract_names(import_) for import_ in tree.iter_imports()])

    # replace using pkg_name mapping and ignore standard lib
    pkgs_final = [
        _PACKAGE_MAPPING.get(name, name) for name in pkgs
        if place_module(name) == 'THIRDPARTY'
    ]

    # parse magics
    leaf = tree.get_first_leaf()

    while leaf:
        if leaf.value == '%' and leaf.get_next_leaf().value == 'load_ext':
            ext_name = leaf.get_next_leaf().get_next_leaf().value
            if ext_name not in _BUILT_IN_EXTENSIONS:
                pkgs_final.append(_PACKAGE_MAPPING.get(ext_name, ext_name))

        leaf = leaf.get_next_leaf()

    # remove duplicates and sort
    return sorted(set(pkgs_final))


def extract_code(nb):
    try:
        code = '\n'.join([
            cell['source'] for cell in nb['cells']
            if cell['cell_type'] == "code"
        ])
    except TypeError:
        # if passed a notebook read using json instead of nbformat
        code = '\n'.join([
            '\n'.join(c['source']) for c in nb['cells']
            if c['cell_type'] == 'code'
        ])

    return code


def to_text(nb):
    try:
        plain = '\n'.join([cell['source'] for cell in nb['cells']])
    except TypeError:
        # # if passed a notebook read using json instead of nbformat
        plain = '\n'.join(['\n'.join(c['source']) for c in nb['cells']])

    return plain


def extract_imports_from_path_to_notebook(path):
    nb = json.loads(Path(path).read_text(encoding='utf-8'))
    return extract_imports_from_notebook(nb)


def extract_imports_from_path_to_script(path):
    code = Path(path).read_text()
    return packages_used(parso.parse(code)) + extract_from_plain_text(code)


def extract_imports_from_notebook(nb):
    code = extract_code(nb)
    # TODO: maybe use the notebook's JSON string? cause here we're ignoring
    # code cells that might have comments such as "pip install something"
    plain = to_text(nb)

    return packages_used(parso.parse(code)) + extract_from_plain_text(plain)


def bootstrap_env(path_to_notebook, name=None, verbose=False):
    # TODO: maybe print "remember to click on save?" - if users edit the
    # notebook but do not save, we won't see their changes
    import nbformat

    if name is None:
        print("name not supplied, installing in the current environment...")

    print('Parsing notebook...')
    nb = nbformat.read(path_to_notebook, as_version=nbformat.NO_CONVERT)
    imports = set(extract_imports_from_notebook(nb)) | {'k2s'}

    cd = ChannelData()
    conda, pip = cd.pkg_exists(imports)

    print(f'Found: {", ".join(imports)}')

    install(name, conda, requirements_pip=pip, verbose=verbose)

    # if installed in a new env, we need to refresh jupyter for the new kernel
    # to appear
    # TODO: only say "switch" to if on a different kernel
    if name is not None:
        print('Kernel is ready! Refresh your browser and switch '
              f'the kernel to: {name}')

    nb.metadata.kernelspec = {
        'display_name': f'Python 3 ({name})',
        'language': 'python',
        'name': name,
    }

    # override kernel spec - note that refreshing won't switch to this
    # kernel and it'll stick with the original one until we kill the
    # existing kernel
    Path(path_to_notebook).write_text(nbformat.v4.writes(nb))



"""
pip install k2s --upgrade && k2s get ploomber/jupysql/doc/nb.ipynb
"""
import shutil
import os
import json
from pathlib import Path, PurePosixPath
from subprocess import run as subprocess_run
import venv
from urllib.request import urlretrieve

from k2s.parse import (extract_imports_from_notebook, download_files,
                       extract_code)

from k2s.subprocess import _run_command


def path_to_bin(env_name, bin):
    dir = 'Scripts' if os.name == 'nt' else 'bin'
    bin_name = f'{bin}.exe' if os.name == 'nt' else bin
    return str(Path(env_name, dir, bin_name))


def from_url(url):
    """

    >>> from k2s.bootstrap import from_url; from_url("URL")
    """
    url = f'https://raw.githubusercontent.com/{url}'

    file = PurePosixPath(url).name
    urlretrieve(url, file)

    return from_file(file, url=url)


def from_file(file, url=None):
    env_name = Path(file).stem + '-env'

    nb = json.loads(Path(file).read_text(encoding='utf-8'))

    if 'metadata' not in nb:
        nb['metadata'] = {}

    if not nb['metadata'].get('kernelspec'):
        print('Notebook is missing kernel information, assuming Python...')
        nb['metadata']['kernelspec'] = {
            "display_name": "Python 3 (ipykernel)",
            "language": "python",
            "name": "python3"
        }

        Path(file).write_text(json.dumps(nb), encoding='utf-8')

    deps = extract_imports_from_notebook(nb)

    # FIXME: we're runnign extract_code inside extract_imports already, we
    # should refactor
    if url:
        deps_extra = download_files(extract_code(nb), url)
    else:
        deps_extra = set()

    deps = list(set(deps) | deps_extra)

    print("Creating virtual environment...")
    print(f"Installing: {', '.join(deps)}")
    print("This may take a moment. "
          "Join our community while you wait: https://ploomber.io/community")

    # if the user stopped the process in the middle and tried again, the env
    # might exist but will be corrupted, so we must clean it
    if Path(env_name).exists():
        shutil.rmtree(env_name)

    # maybe site-packages, yes? will that copy the existing ones? perhaps
    # they already ahve most of the dependencies and we can save time.
    # if so, then we need to ensure jupyterlab starts with the kernel in
    # the virtual environment
    venv.create(env_name, with_pip=True, system_site_packages=False)

    path_to_python = path_to_bin(env_name, 'python')

    subprocess_run(
        [path_to_python, '-m', 'pip', 'install', 'pip', '--upgrade'],
        check=True,
        capture_output=True)

    _run_command([path_to_python, '-m', 'pip', 'install', 'jupyterlab'] + deps)

    print("Lauching Jupyter... (this might take a few seconds)")

    # TODO: if jupyter lab is already installed, we can use it (and skip
    # installation), but we have to make sure that the kernel used is
    # the one in the venv, not in the parent env
    path_to_jupyterlab = path_to_bin(env_name, 'jupyter-lab')

    print('To exit: CTRL + C')

    try:
        subprocess_run([path_to_jupyterlab, file],
                       check=True,
                       capture_output=True)
    except KeyboardInterrupt:
        print('Exiting...')
        print(f'\nTo laucnh again:\n\n'
              f'{path_to_jupyterlab} {file}')



class KernelRestartRequired(Exception):

    def __init__(self):
        msg = ("Successful conda installation, restarting kernel. "
               "Please re-run this cell to continue.")
        super().__init__(msg)


class CLIError(Exception):
    pass



import sys
from os import environ
from subprocess import Popen, PIPE

from k2s.exceptions import CLIError
from k2s.spinner import Spinner

DEBUG_MODE = bool(environ.get("K2S_DEBUG", False))


def _run_command(cmd):
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)

    last_length = 0

    with Spinner():
        while True:
            # TODO: this wont work if the output is larger than the terminal
            # size
            output = process.stdout.readline().decode()

            if output == '' and process.poll() is not None:
                break

            if output:
                line = output.strip() + '  '

                if DEBUG_MODE:
                    print(line)
                else:
                    out = ('\r' + ' ' * last_length + '\r' + line)
                    sys.stdout.write(out)
                    sys.stdout.flush()
                    last_length = len(line)

    print('\n', end='')

    return_code = process.poll()

    if return_code:
        raise CLIError(
            'An error happened:\n\n'
            f'{process.stderr.read().decode()}\nFor help, share this error '
            'in the #ask-anything channel in '
            'our community: https://ploomber.io/community')
